# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import os
import tempfile

import cliapp

import vmdb


class AnsiblePlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(AnsibleStepRunner())


class AnsibleStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['ansible', 'playbook']

    def run(self, step, settings, state):
        fstag = step['ansible']
        playbook = step['playbook']
        mount_point = state.mounts[fstag]

        vmdb.progress(
            'Running ansible playbook {} on filesystem at {} ({})'.format(
                playbook, mount_point, fstag))

        state.ansible_inventory = self.create_inventory(mount_point)
        vmdb.progress(
            'Created {} for Ansible inventory'.format(state.ansible_inventory))

        env = dict(os.environ)
        env['ANSIBLE_NOCOWS'] = '1'
        vmdb.runcmd(
            ['ansible-playbook', '-c', 'chroot',
             '-i', state.ansible_inventory, playbook],
            env=env)

    def teardown(self, step, settings, state):
        if hasattr(state, 'ansible_inventory'):
            vmdb.progress('Removing {}'.format(state.ansible_inventory))
            os.remove(state.ansible_inventory)

    def create_inventory(self, chroot):
        fd, filename = tempfile.mkstemp()
        os.write(fd, '[image]\n{}\n'.format(chroot).encode())
        os.close(fd)
        return filename



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import os
import tempfile

import cliapp

import vmdb


class MountPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(MountStepRunner())


class MountStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['mount', 'fs-tag']

    def run(self, step, settings, state):
        self.mount_rootfs(step, settings, state)

    def teardown(self, step, settings, state):
        self.unmount_rootfs(step, settings, state)

    def mount_rootfs(self, step, settings, state):
        if not hasattr(state, 'mounts'):
            state.mounts = {}

        part_tag = step['mount']
        fs_tag = step['fs-tag']
        dirname = step.get('dirname')
        mount_on = step.get('mount-on')

        if fs_tag in state.mounts:
            raise Exception('fs-tag {} already used'.format(fs_tag))

        if dirname:
            if not mount_on:
                raise Exception('no mount-on tag given')

            if mount_on not in state.mounts:
                raise Exception('cannot find tag {}'.format(mount_on))

            mount_point = os.path.join(
                state.mounts[mount_on], './' + step['dirname'])

            if not os.path.exists(mount_point):
                os.makedirs(mount_point)
        else:
            mount_point = tempfile.mkdtemp()

        device = state.parts[part_tag]

        vmdb.runcmd(['mount', device, mount_point])
        state.mounts[fs_tag] = mount_point

        return mount_point

    def unmount_rootfs(self, step, settings, state):
        fs_tag = step['fs-tag']
        mount_point = state.mounts[fs_tag]

        vmdb.runcmd(['umount', mount_point])
        if not step.get('mount-on'):
            os.rmdir(mount_point)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


class State(object):

    def __init__(self):
        self._attrs = {}  # make sure this attribute exists
        self._attrs = self.as_dict()

    def as_dict(self):
        return {
            key: getattr(self, key)
            for key in dir(self)
            if not key in self._attrs
        }



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import logging

import cliapp

import vmdb


class EchoPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(EchoStepRunner())


class EchoStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['echo']

    def run(self, step, settings, state):
        text = step['echo']
        vmdb.progress('{}'.format(text))

    def teardown(self, step, settings, state):
        if 'teardown' in step:
            text = step['teardown']
            vmdb.progress('{}'.format(text))
            logging.info('%s', text)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import os
import sys

import cliapp


_verbose = False


def set_verbose_progress(verbose):
    global _verbose
    _verbose = verbose


def error(msg):
    logging.error(msg, exc_info=True)
    if _verbose:
        sys.stderr.write('ERROR: {}\n'.format(msg))


def progress(msg):
    logging.info(msg)
    if _verbose:
        sys.stdout.write('{}\n'.format(msg))


def runcmd(argv, *argvs, **kwargs):
    progress('Exec: %r' % (argv,))
    kwargs['stdout_callback'] = _log_stdout
    kwargs['stderr_callback'] = _log_stderr
    env = kwargs.get('env', os.environ.copy())
    env['LC_ALL'] = 'C'
    kwargs['env'] = env
    return cliapp.runcmd(argv, *argvs, **kwargs)


def runcmd_chroot(chroot, argv, *argvs, **kwargs):
    full_argv = ['chroot', chroot] + argv
    return runcmd(full_argv, *argvs, **kwargs)


def _log_stdout(data):
    logging.debug('STDOUT: %r', data)
    return data


def _log_stderr(data):
    logging.debug('STDERR: %r', data)
    return data



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import os

import cliapp

import vmdb


class ChrootPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(ChrootStepRunner())
        self.app.step_runners.add(ShellStepRunner())


class ChrootStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['chroot', 'shell']

    def run(self, step, settings, state):
        fs_tag = step['chroot']
        shell = step['shell']

        mount_point = state.mounts[fs_tag]

        vmdb.progress(
            'chroot {} to {}'.format(mount_point, ' '.join(shell.split('\n'))))
        vmdb.runcmd_chroot(mount_point, ['sh', '-c', shell])


class ShellStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['shell', 'root-fs']

    def run(self, step, settings, state):
        shell = step['shell']
        fs_tag = step['root-fs']

        vmdb.progress(
            'run shell {}'.format(' '.join(shell.split('\n'))))
        env = dict(os.environ)
        env['ROOT'] = state.mounts[fs_tag]
        vmdb.runcmd(['sh', '-c', shell], env=env)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import os

import cliapp

import vmdb


class RootFSCachePlugin(cliapp.Plugin):

    def enable(self):
        self.app.settings.string(
            ['rootfs-tarball'],
            'store rootfs cache tar archives in FILE',
            metavar='FILE')

        self.app.step_runners.add(MakeCacheStepRunner())
        self.app.step_runners.add(UnpackCacheStepRunner())


class MakeCacheStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['cache-rootfs']

    def run(self, step, settings, state):
        fs_tag = step['cache-rootfs']
        rootdir = state.mounts[fs_tag]
        tar_path = settings['rootfs-tarball']
        if not tar_path:
            raise Exception('--rootfs-tarball MUST be set')
        if not os.path.exists(tar_path):
            vmdb.progress(
                'Caching contents of {} to {}'.format(rootdir, tar_path))
            vmdb.runcmd(['tar', '-C', rootdir, '--one-file-system',
                         '-caf', tar_path, '.'])


class UnpackCacheStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['unpack-rootfs']

    def run(self, step, settings, state):
        fs_tag = step['unpack-rootfs']
        rootdir = state.mounts[fs_tag]
        tar_path = settings['rootfs-tarball']
        if not tar_path:
            raise Exception('--rootfs-tarball MUST be set')
        if os.path.exists(tar_path):
            vmdb.progress(
                'Unpacking rootfs from {} to {}'.format(tar_path, rootdir))
            vmdb.runcmd(
                ['tar', '-C', rootdir, '-xf', tar_path, '--numeric-owner'])
            state.rootfs_unpacked = True



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


from .version import __version__, __version_info__
from .state import State
from .step_list import (
    StepRunnerList,
    StepRunnerInterface,
    NoMatchingRunner,
    StepError,
)
from .runcmd import (
    runcmd,
    runcmd_chroot,
    set_verbose_progress,
    progress,
    error,
)
from .spec import (
    Spec,
    expand_templates,
)
from .app import Vmdb2



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import unittest

import vmdb


class StepRunnerListTests(unittest.TestCase):

    def test_is_empty_initially(self):
        steps = vmdb.StepRunnerList()
        self.assertEqual(len(steps), 0)

    def test_adds_a_runner(self):
        steps = vmdb.StepRunnerList()
        runner = DummyStepRunner()
        steps.add(runner)
        self.assertEqual(len(steps), 1)

    def test_finds_correct_runner(self):
        steps = vmdb.StepRunnerList()
        runner = DummyStepRunner()
        steps.add(runner)
        found = steps.find({'foo': None, 'bar': None})
        self.assertEqual(runner, found)

    def test_raises_error_if_runner_not_found(self):
        steps = vmdb.StepRunnerList()
        runner = DummyStepRunner()
        steps.add(runner)
        with self.assertRaises(vmdb.NoMatchingRunner):
            steps.find({'foo': None})


class DummyStepRunner(vmdb.StepRunnerInterface):

    def run(self, *args):
        pass

    def get_required_keys(self):
        return ['foo', 'bar']



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import cliapp

import vmdb


class PartitionPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(MklabelStepRunner())
        self.app.step_runners.add(MkpartStepRunner())


class MklabelStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['mklabel']

    def run(self, step, settings, state):
        label_type = step['mklabel']
        device = step['device']
        vmdb.progress(
            'Creating partition table ({}) on {}'.format(label_type, device))
        vmdb.runcmd(['parted', '-s', device, 'mklabel', label_type])
        state.parts = {}


class MkpartStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['mkpart']

    def run(self, step, settings, state):
        part_type = step['mkpart']
        device = step['device']
        start = step['start']
        end = step['end']
        part_tag = step['part-tag']
        fs_type = step.get('fs-type', 'ext2')

        vmdb.progress(
            'Creating partition ({}) on {} ({} to {})'.format(
                part_type, device, start, end))
        vmdb.runcmd(['parted', '-s', device, 'mkpart', part_type, fs_type, start, end])

        vmdb.runcmd(['kpartx', '-dsv', device])
        output = vmdb.runcmd(['kpartx', '-asv', device]).decode('UTF-8')
        device_file = None
        for line in output.splitlines():
            words = line.split()
            if words[0] == 'add':
                name = words[2]
                device_file = '/dev/mapper/{}'.format(name)

        assert device_file is not None
        parts = getattr(state, 'parts', {})
        parts[part_tag] = device_file
        state.parts = parts

    def teardown(self, step, settings, state):
        device = step['device']
        vmdb.progress(
            'Undoing loopback devices for partitions on {}'.format(device))
        vmdb.runcmd(['kpartx', '-dsv', device])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import cliapp


class StepRunnerInterface(object):  # pragma: no cover

    def get_required_keys(self):
        raise NotImplementedError()

    def run(self, step_spec, settings, state):
        raise NotImplementedError()

    def teardown(self, step_spec, settings, state):
        # Default implementation does nop, so that sub-classes don't
        # need to have a nop teardown.
        pass

    def skip(self, step_spec, settings, state):
        # Return true if step should be skipped and not run. Does not
        # apply to teardowns.

        # Skipping is indicated by the step having a field 'unless',
        # which is either the name of a variable (field in state), or
        # a list of such names. If all variables have a value that
        # evaluates as truth, the step is skipped.

        value = step_spec.get('unless', None)
        if value is None:
            return False

        if isinstance(value, list):
            return all(getattr(state, field, False) for field in value)
        return getattr(state, value, False)


class StepRunnerList(object):

    def __init__(self):
        self._runners = []

    def __len__(self):
        return len(self._runners)

    def add(self, runner):
        self._runners.append(runner)

    def find(self, step_spec):
        actual = set(step_spec.keys())
        for runner in self._runners:
            required = set(runner.get_required_keys())
            if actual.intersection(required) == required:
                return runner
        raise NoMatchingRunner(actual)


class StepError(cliapp.AppException):

    pass


class NoMatchingRunner(cliapp.AppException):

    def __init__(self, keys):
        super(NoMatchingRunner, self).__init__(
            'No runner implements step with keys {}'.format(
                ', '.join(keys)))



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import os

import cliapp

import vmdb


class AptPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(AptStepRunner())


class AptStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['apt', 'fs-tag', 'packages']

    def run(self, step, settings, state):
        operation = step['apt']
        if operation != 'install':
            raise Exception('"apt" must always have value "install"')

        packages = step['packages']
        fstag = step['fs-tag']
        mount_point = state.mounts[fstag]

        if not self.got_eatmydata(state):
            self.install_packages(mount_point, [], ['eatmydata'])
            state.got_eatmydata = True
        self.install_packages(mount_point, ['eatmydata'], packages)

    def got_eatmydata(self, state):
        return hasattr(state, 'got_eatmydata') and getattr(state, 'got_eatmydata')

    def install_packages(self, mount_point, argv_prefix, packages):
        env = os.environ.copy()
        env['DEBIAN_FRONTEND'] = 'noninteractive'

        vmdb.runcmd_chroot(
            mount_point,
            argv_prefix +
            ['apt-get', '-y', '--no-show-progress', 'install'] + packages,
            env=env)



# Copyright 2017  Lars Wirzenius and Stuart Prescott
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import cliapp

import vmdb


class QemuDebootstrapPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(QemuDebootstrapStepRunner())


class QemuDebootstrapStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['qemu-debootstrap', 'target', 'mirror', 'arch']

    def run(self, step, settings, state):
        suite = step['qemu-debootstrap']
        tag = step['target']
        target = state.mounts[tag]
        mirror = step['mirror']
        variant = step.get('variant', '-')
        arch = step['arch']
        components = step.get('components', ['main'])
        if not (suite and tag and target and mirror and arch):
            raise Exception('missing arg for qemu-debootstrap step')
        vmdb.progress(
            'Qemu-debootstrap {} {} {} {}'.format(suite, target, mirror, arch))
        vmdb.runcmd(
            ['qemu-debootstrap',
             '--arch', arch,
             '--variant', variant,
             '--components', ','.join(components), suite,
             target,
             mirror])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import logging
import os

import cliapp

import vmdb


class VirtualFilesystemMountPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(VirtualFilesystemMountStepRunner())


class VirtualFilesystemMountStepRunner(vmdb.StepRunnerInterface):

    virtuals = [
        ['none', '/proc', 'proc'],
        ['none', '/dev', 'devtmpfs'],
        ['none', '/dev/pts', 'devpts'],
        ['none', '/dev/shm', 'tmpfs'],
        ['none', '/run', 'tmpfs'],
        ['none', '/run/lock', 'tmpfs'],
        ['none', '/sys', 'sysfs'],
    ]

    def get_required_keys(self):
        return ['mount-virtual-filesystems']

    def run(self, step, settings, state):
        fstag = step['mount-virtual-filesystems']
        mount_point = state.mounts[fstag]
        self.mount_virtuals(mount_point, state)

    def teardown(self, step, settings, state):
        self.unmount_virtuals(state)

    def mount_virtuals(self, rootfs, state):
        if not hasattr(state, 'virtuals'):
            state.virtuals = []

        for device, mount_point, fstype in self.virtuals:
            path = os.path.join(rootfs, './' + mount_point)
            if not os.path.exists(path):
                os.mkdir(path)
            vmdb.runcmd(['mount', '-t', fstype, device, path])
            state.virtuals.append(path)
        logging.debug('mounted virtuals: %r', state.virtuals)

    def unmount_virtuals(self, state):
        logging.debug('unmounting virtuals: %r', state.virtuals)
        for mount_point in reversed(state.virtuals):
            try:
                vmdb.runcmd(['umount', mount_point])
            except cliapp.AppException:
                vmdb.error('Something went wrong while unmounting. Ignoring.')



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import cliapp

import vmdb


class MkfsPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(MkfsStepRunner())


class MkfsStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['mkfs', 'partition']

    def run(self, step, settings, state):
        fstype = step['mkfs']
        part_tag = step['partition']
        device = state.parts[part_tag]
        vmdb.progress(
            'Creating {} filesystem on {}'.format(fstype, device))
        vmdb.runcmd(['/sbin/mkfs', '-t', fstype, device])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import sys

import cliapp

import vmdb


class Vmdb2(cliapp.Application):

    def add_settings(self):
        self.settings.string(
            ['image'],
            'create image file FILE',
            metavar='FILE')

        self.settings.boolean(
            ['verbose', 'v'],
            'verbose output')

    def setup(self):
        self.step_runners = vmdb.StepRunnerList()

    def process_args(self, args):
        if len(args) != 1:
            sys.exit("No image specification was given on the command line.")

        vmdb.set_verbose_progress(self.settings['verbose'])

        spec = self.load_spec_file(args[0])
        state = vmdb.State()
        params = self.create_template_vars(state)
        steps = spec.get_steps(params)

        # Check that we have step runners for each step
        for step in steps:
            self.step_runners.find(step)

        steps_taken, core_meltdown = self.run_steps(steps, state)
        if core_meltdown:
            vmdb.progress('Something went wrong, cleaning up!')
        else:
            vmdb.progress('All went fine, cleaning up.')
        self.run_teardowns(steps_taken, state)

        if core_meltdown:
            logging.error('An error occurred, exiting with non-zero exit code')
            sys.exit(1)

    def load_spec_file(self, filename):
        vmdb.progress('Load spec file {}'.format(filename))
        spec = vmdb.Spec()
        spec.load_file(filename)
        return spec

    def run_steps(self, steps, state):
        return self.run_steps_helper(
            steps, state, 'Running step: %r', 'run', False)

    def run_teardowns(self, steps, state):
        return self.run_steps_helper(
            list(reversed(steps)), state, 'Running teardown: %r', 'teardown', True)

    def run_steps_helper(self, steps, state, msg, method_name, keep_going):
        core_meltdown = False
        steps_taken = []

        for step in steps:
            try:
                logging.info(msg, step)
                steps_taken.append(step)
                runner = self.step_runners.find(step)
                if runner.skip(step, self.settings, state):
                    logging.info('Skipping as requested')
                else:
                    method = getattr(runner, method_name)
                    method(step, self.settings, state)
            except BaseException as e:
                vmdb.error(str(e))
                core_meltdown = True
                if not keep_going:
                    break

        return steps_taken, core_meltdown

    def create_template_vars(self, state):
        vars = dict()
        for key in self.settings:
            vars[key] = self.settings[key]
        vars.update(state.as_dict())
        return vars



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=



import cliapp

import vmdb


class ErrorPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(ErrorStepRunner())


class ErrorStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['error', 'teardown']

    def run(self, step, settings, state):
        # We use vmdb.progress here to get output to go to stdout,
        # instead of stderr. We want that for tests.
        vmdb.progress('{}'.format(step['error']))
        raise vmdb.StepError('an error occurred')

    def teardown(self, step, settings, state):
        # We use vmdb.progress here to get output to go to stdout,
        # instead of stderr. We want that for tests.
        vmdb.progress('{}'.format(step['teardown']))



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import io
import os
import tempfile
import unittest

import yaml

import vmdb


class SpecTests(unittest.TestCase):

    spec_yaml = '''
    steps:
      - step: foo
        arg: "{{ var1 }}"
      - step: bar
    '''

    def setUp(self):
        self.filename = write_temp_file(bytes(self.spec_yaml, 'ascii'))
        self.spec = vmdb.Spec()

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_loads_spec(self):
        self.spec.load_file(self.filename)
        self.assertEqual(self.spec.as_dict(), as_dict(self.spec_yaml))

    def test_expands_templates(self):
        self.spec.load_file(self.filename)
        params = {
            'var1': 'value1',
        }
        steps = self.spec.get_steps(params)
        self.assertEqual(
            steps,
            [
                {
                    'step': 'foo',
                    'arg': 'value1',
                },
                {
                    'step': 'bar',
                },
            ]
        )

class ExpandTemplatesTests(unittest.TestCase):

    def test_raises_assert_if_given_incomprehensible_value(self):
        with self.assertRaises(AssertionError):
            vmdb.expand_templates(None, {})

    def test_returns_same_given_string_without_template(self):
        self.assertEqual(vmdb.expand_templates('foo', {}), 'foo')

    def test_expands_simple_string_template(self):
        params = {
            'foo': 'bar',
        }
        self.assertEqual(vmdb.expand_templates('{{ foo }}', params), 'bar')

    def test_expands_list_of_templates(self):
        params = {
            'foo': 'bar',
        }
        self.assertEqual(vmdb.expand_templates(['{{ foo }}'], params), ['bar'])

    def test_expands_dict_of_templates(self):
        params = {
            'foo': 'bar',
        }
        self.assertEqual(
            vmdb.expand_templates({'key': '{{ foo }}'}, params),
            {'key': 'bar'}
        )


def write_temp_file(data):
    fd, filename = tempfile.mkstemp()
    os.write(fd, data)
    os.close(fd)
    return filename


def as_dict(yaml_text):
    with io.StringIO(yaml_text) as f:
        return yaml.safe_load(f)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


# Installing GRUB onto a disk image is a bit of a black art. I haven't
# found any good documentation for it. This plugin is written based on
# de-ciphering the build_openstack_image script. Here is an explanation
# of what I _THINK_ is happening.
#
# The crucial command is grub-install. It needs a ton of options to
# work correctly: see below in the code for the list, and the manpage
# for an explanation of what each of them means. We will be running
# grub-install in a chroot so that we use the version in the Debian
# version we're installing, rather than the host system, which might
# be any Debian version.
#
# To run grub-install in a chroot, we need to set up the chroot in
# various ways. Firstly, we need to tell grub-install which device
# file the image has. We can't just give it the image file itself,
# since it isn't inside the chroot, so instead we arrange to have a
# loop block device that covers the whole image file, and we bind
# mount /dev into the chroot so the device is available.
#
# grub-install seems to also require /proc and /sys so we bind mount
# those into the chroot as well.
#
# We install the UEFI version of GRUB, and for that we additionally
# bind mount the EFI partition in the image. Oh yeah, you MUST have
# one.
#
# We also make sure the right GRUB package is installed in the chroot,
# before we run grub-install.
#
# Further, there's some configuration tweaking we need to do. See the
# code. Don't ask me why they're necessary.
#
# For cleanliness, we also undo any bind mounts into the chroot. Don't
# want to leave them in case they cause trouble.
#
# Note that this is currently rather strongly assuming that UEFI and
# the amd64 (a.k.a. x86_64) architecture are being used. These should
# probably not be hardcoded. Patch welcome.

# To use this plugin: write steps to create a root filesystem, and an
# VFAT filesystem to be mounted as /boot/efi. Install Debian onto the
# root filesystem. Then install grub with a step like this:
#
#         - grub: uefi
#           root-fs: root-fs
#           root-part: root-part
#           efi-part: efi-part
#
# Here: root-fs is the tag for the root filesystem, root-part is the
# tag for the partition with the root filesystem, and efi-part is tag
# for the EFI partition.
#
# The grub step will take of the rest.


import os
import re

import cliapp

import vmdb


class GrubPlugin(cliapp.Plugin):

    def enable(self):
        self.app.step_runners.add(GrubStepRunner())


class GrubStepRunner(vmdb.StepRunnerInterface):

    def get_required_keys(self):
        return ['grub', 'root-fs']

    def run(self, step, settings, state):
        flavor = step['grub']
        if flavor == 'uefi':
            self.install_uefi(step, settings, state)
        elif flavor == 'bios':
            self.install_bios(step, settings, state)
        else:
            raise Exception('Unknown GRUB flavor {}'.format(flavor))

    def install_uefi(self, step, settings, state):
        vmdb.progress('Installing GRUB for UEFI')

        if not 'efi-part' in step:
            raise Exception('"efi-part" is required in UEFI GRUB installtion')

        grub_package = 'grub-efi-amd64'
        grub_target = 'x86_64-efi'
        self.install_grub(step, settings, state, grub_package, grub_target)

    def install_bios(self, step, settings, state):
        vmdb.progress('Installing GRUB for BIOS')

        grub_package = 'grub-pc'
        grub_target = 'i386-pc'
        self.install_grub(step, settings, state, grub_package, grub_target)

    def install_grub(self, step, settings, state, grub_package, grub_target):
        console = step.get('console', None)

        rootfs = step['root-fs']
        chroot = state.mounts[rootfs]

        root_part = step['root-part']
        root_dev = state.parts[root_part]

        image_dev = self.get_image_loop_device(root_dev)

        if 'efi-part' in step:
            efi_part = step['efi-part']
            efi_dev = state.parts[efi_part]
        else:
            efi_dev = None

        self.bind_mount_many(chroot, ['/dev', '/proc', '/sys'], state)
        if efi_dev:
            self.mount(chroot, efi_dev, '/boot/efi', state)
        self.install_package(chroot, grub_package)

        kernel_params = [
            'biosdevname=0',
            'net.ifnames=0',
            'consoleblank=0',
            'systemd.show_status=true',
            'rw',
            'quiet',
            'systemd.show_status=false',
        ]
        if console == 'serial':
            kernel_params.extend([
                'quiet',
                'loglevel=3',
                'rd.systemd.show_status=false',
                'systemd.show_status=false',
                'console=tty0',
                'console=ttyS0,115200n8',
            ])

        self.set_grub_cmdline_config(chroot, kernel_params)
        if console == 'serial':
            self.add_grub_serial_console(chroot)

        vmdb.runcmd_chroot(chroot, ['grub-mkconfig', '-o', '/boot/grub/grub.cfg'])
        vmdb.runcmd_chroot(
            chroot, [
                'grub-install',
                '--target=' + grub_target,
                '--no-nvram',
                '--force-extra-removable',
                '--no-floppy',
                '--modules=part_msdos part_gpt',
                '--grub-mkdevicemap=/boot/grub/device.map',
                image_dev,
            ]
        )

        self.unmount(state)

    def teardown(self, step, settings, state):
        self.unmount(state)

    def unmount(self, state):
        mounts = getattr(state, 'grub_mounts', [])
        mounts.reverse()
        while mounts:
            mount_point = mounts.pop()
            vmdb.runcmd(['umount', mount_point])

    def get_image_loop_device(self, partition_device):
        # We get /dev/mappers/loopXpY and return /dev/loopX
        assert partition_device.startswith('/dev/mapper/loop')

        m = re.match(r'^/dev/mapper/(?P<loop>loop\d+)p\d+$', partition_device)
        assert m is not None

        loop = m.group('loop')
        return '/dev/{}'.format(loop)

    def bind_mount_many(self, chroot, paths, state):
        for path in paths:
            self.mount(chroot, path, path, state, mount_opts=['--bind'])

    def mount(self, chroot, path, mount_point, state, mount_opts=None):
        chroot_path = self.chroot_path(chroot, mount_point)
        if not os.path.exists(chroot_path):
            os.makedirs(chroot_path)

        if mount_opts is None:
            mount_opts = []

        vmdb.runcmd(['mount'] + mount_opts + [path, chroot_path])

        binds = getattr(state, 'grub_mounts', None)
        if binds is None:
            binds = []
        binds.append(chroot_path)
        state.grub_mounts = binds

    def chroot_path(self, chroot, path):
        return os.path.normpath(os.path.join(chroot, '.' + path))

    def install_package(self, chroot, package):
        vmdb.progress('Install {} in chroot {}'.format(package, chroot))
        env = os.environ.copy()
        env['DEBIAN_FRONTEND'] = 'noninteractive'
        vmdb.runcmd_chroot(
            chroot,
            ['apt-get', '-y', '--no-show-progress', 'install', package],
            env=env)

    def set_grub_cmdline_config(self, chroot, kernel_params):
        param_string = ' '.join(kernel_params)

        filename = self.chroot_path(chroot, '/etc/default/grub')

        with open(filename) as f:
            text = f.read()

        lines = text.splitlines()
        lines = [line for line in lines
                 if not line.startswith('GRUB_CMDLINE_LINUX_DEFAULT')]
        lines.append('GRUB_CMDLINE_LINUX_DEFAULT="{}"'.format(param_string))

        with open(filename, 'w') as f:
            f.write('\n'.join(lines) + '\n')

    def add_grub_serial_console(self, chroot):
        filename = self.chroot_path(chroot, '/etc/default/grub')

        with open(filename, 'a') as f:
            f.write('GRUB_TERMINAL=serial\n')
            f.write('GRUB_SERIAL_COMMAND="serial --speed=115200 --unit=0 '
                    '--word=8 --parity=no --stop=1"\n')



#!/usr/bin/env python3
# Copyright (C) 2017  Lars Wirzenius <liw@liw.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from distutils.core import setup, Extension
from distutils.cmd import Command
from distutils.command.build import build
from distutils.command.clean import clean
import os
import glob

import cliapp

import vmdb


class Build(build):

    def run(self):
        build.run(self)
        self.build_manpage('vmdb2', '')
        self.format_yarns()

    def build_manpage(self, program, lang):
        return
        # building manpage fails on by unstable CI worker for mysterious
        # reasons, will re-enable later
        print('building manpage for %s (lang=%s)' % (program, lang))
        self.generate_troff(program, lang)
        self.format_man_as_txt(program)

    def generate_troff(self, program, lang):
        with open('%s.1%s' % (program, lang), 'w') as f:
            cliapp.runcmd(
                ['python', program,
                 '--generate-manpage=%s.1%s.in' % (program, lang),
                 '--output=%s.1' % program],
                stdout=f)

    def format_man_as_txt(self, program):
        env = dict(os.environ)
        env['MANWIDTH'] = '80'
        with open('%s.1.txt' % program, 'w') as f:
            cliapp.runcmd(
                ['man', '-l', '%s.1' % program],
                ['col', '-b'],
                stdout=f,
                env=env)

    def format_yarns(self):
        print('building yarns')
        cliapp.runcmd(['make', '-C', 'yarns'])


setup(
    name='vmdb2',
    version=vmdb.__version__,
    description='create disk image with Debian installed',
    author='Lars Wirzenius',
    author_email='liw@liw.fi',
    url='http://liw.fi/vmdebootstrap/',
    scripts=['vmdb2'],
    packages=['vmdb', 'vmdb.plugins'],
    data_files=[('share/man/man1', glob.glob('*.1'))],
    cmdclass={
        'build': Build,
    },
)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import jinja2
import yaml


class Spec:

    def __init__(self):
        self._dict = None

    def load_file(self, filename):
        with open(filename) as f:
            self._dict = yaml.safe_load(f)

    def as_dict(self):
        return dict(self._dict)

    def get_steps(self, params):
        return expand_templates(self._dict['steps'], params)


def expand_templates(value, params):
    if isinstance(value, str):
        template = jinja2.Template(value)
        return template.render(**params)
    elif isinstance(value, list):
        return [expand_templates(x, params) for x in value]
    elif isinstance(value, dict):
        return {
            key: expand_templates(value[key], params)
            for key in value
        }
    else:
        assert 0, 'Unknown value type: {!r}'.format(value)
